#!/usr/bin/perl -w
# Filename:	reslog.pl
# Description:	Perl script to resolve apache log files
# Author:	imacat <imacat@mail.imacat.idv.tw>
# Date:		2000-12-22
# Copyright:	(c) 2000,2001 imacat

use strict;
use Fcntl qw(:flock);
use Socket;
use IO::Handle qw(autoflush);

use vars qw($LOG_DIR $DEFAULT_SUFFIX);
use vars qw($DEBUG $USETMP);
*LOG_DIR = \"/usr/local/apache/logs";
*DEFAULT_SUFFIX = \".resolved";

use vars qw($THIS_FILE $SHORTHELP $VERNO $VERSTR $HELPSTR);
$0 =~ /[^\/]*$/;
$THIS_FILE = $&;
$SHORTHELP = "Try `$THIS_FILE --help' for more information.";
$VERNO = "1.3.1";
$VERSTR = "$& v$VERNO written by imacat <imacat\@mail.imacat.idv.tw>.\n";
$HELPSTR = << "EOF";
Usage: $THIS_FILE [-dhtv] [-s suffix] [logfile]
Resolve IPs from apache access log <logfile> and append the result to
<logfile>$DEFAULT_SUFFIX for later use by the analog program.

  -d,--debug        Display debug messages.
  -h,--help         Display this help.
  -t,--usetmp       use the last unfinished temporary working file as source.
  -s,--suffix .suf  Use the suffix .suf, instead of the default suffix $DEFAULT_SUFFIX,
                    on the resolved file.
  -v,--version      Display version number.
  logfile           The apache access log file to be resolved. If a relative
                    path is specified, it is related to the
                    $LOG_DIR by default. If the <logfile> is a
                    directory, it will look for the file <logfile>/access_log.

It will copy the <logfile> to a temporary working file <logfile>.tmp and
restart the <logfile> first. Then it will resolve the <logfile>.tmp. The
result will be appended to the <logfile>$DEFAULT_SUFFIX and the temporary
<logfile>.tmp will be removed. If this program crashes during execution,
leaving an unfinished <logfile>.tmp, you can specify -t to force resolving
the <logfile>.tmp.

EOF

main();
exit;

# D{
sub main {
	local $_;
	my ($source, $result, $ip, $dns, %dns, $rec_counter, $ip_counter, $dns_counter);
	
	($source, $result) = parse_args(@ARGV);
	$| = 1 if (defined $DEBUG);
	
	if (defined $USETMP) {
		die "$THIS_FILE: Temporary working file does not exist.\n"
			unless (-e "$source.tmp");
	} else {
		die "$THIS_FILE: Temporary working file exists. Program is possibily running now.\n"
			if (-e "$source.tmp");
		if (-z $source) {
			open (DATA_TMP, ">$source.tmp") || die "$THIS_FILE: $!";
			close (DATA_TMP) || die "$THIS_FILE: $!";
		} else {
			open (DATA, "+<$source") || die "$THIS_FILE: $!";
			flock (DATA, LOCK_EX) || die "$THIS_FILE: $!";
			open (DATA_TMP, ">$source.tmp") || die "$THIS_FILE: $!";
			flock (DATA_TMP, LOCK_EX) || die "$THIS_FILE: $!";
			(@_ = <DATA>) || die "$THIS_FILE: $!";
			$_ = join("", @_);
			(print DATA_TMP) || die "$THIS_FILE: $!";
			flock (DATA_TMP, LOCK_UN) || die "$THIS_FILE: $!";
			close (DATA_TMP) || die "$THIS_FILE: $!";
			seek (DATA, 0, 0) || die "$THIS_FILE: $!";
			truncate (DATA, 0) || die "$THIS_FILE: $!";
			flock (DATA, LOCK_UN) || die "$THIS_FILE: $!";
			close (DATA) || die "$THIS_FILE: $!";
		}
	}
	
	if (-z "$source.tmp") {
		@_ = qw();
	} else {
		open (DATA, "$source.tmp") || die "$THIS_FILE: $!";
		flock (DATA, LOCK_SH) || die "$THIS_FILE: $!";
		(@_ = <DATA>) || die "$THIS_FILE: $!";
		flock (DATA, LOCK_UN) || die "$THIS_FILE: $!";
		close (DATA) || die "$THIS_FILE: $!";
	}
	$rec_counter = @_;
	print "$source $rec_counter records read.\n" if (defined $DEBUG);
	
	$ip_counter = 0;
	$dns_counter = 0;
	for ($_ = 0; $_ < @_; $_++) {
		$_[$_] =~ /^\S+/;
		$ip = $&;
		if ($ip !~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
		} elsif (exists $dns{$ip}) {
			$_[$_] =~ s/^\S+/$dns{$ip}/;
		} else {
			$ip_counter++;
			print "[" . ($_ + 1) . "/$rec_counter] Looking for [$ip] ... " if (defined $DEBUG);
			$dns = gethostbyaddr(inet_aton($ip), AF_INET);
			if (defined $dns) {
				$dns_counter++;
				print "[$dns].\n" if (defined $DEBUG);
			} else {
				print "Failed.\n" if (defined $DEBUG);
				$dns = $ip;
			}
			$dns{$ip} = $dns;
			$_[$_] =~ s/^\S+/$dns/;
		}
	}
	
	$_ = join("", @_);
	open (DATA, ">>$result") || die "$THIS_FILE: $!";
	flock (DATA, LOCK_EX) || die "$THIS_FILE: $!";
	(print DATA) || die "$THIS_FILE: $!";
	flock (DATA, LOCK_UN) || die "$THIS_FILE: $!";
	close (DATA) || die "$THIS_FILE: $!";
	unlink ("$source.tmp") || die "$THIS_FILE: $!";
	print "$ip_counter IPs found, $dns_counter names resolved, " . (time() - $^T) . " seconds elapsed.\n" if (defined $DEBUG);
	return;
}

# parse_args: parse arguments
sub parse_args {
	local $_;
	my (@ARGV, $argn, $c, $source, $result, $suffix);
	@ARGV = @_;
	
	for ($argn = 0; $argn < @ARGV; $argn++) {
		if ($ARGV[$argn] =~ /^--/) {
			if ($ARGV[$argn] eq "--debug") {
				die "$THIS_FILE: $ARGV[$argn] declared more than once.\n$SHORTHELP\n"
					if (defined $DEBUG);
				$DEBUG = 1;
			} elsif ($ARGV[$argn] eq "--usetmp") {
				die "$THIS_FILE: $ARGV[$argn] declared more than once.\n$SHORTHELP\n"
					if (defined $USETMP);
				$USETMP = 1;
			} elsif ($ARGV[$argn] eq "--suffix") {
				die "$THIS_FILE: $ARGV[$argn] declared more than once.\n$SHORTHELP\n"
					if (defined $suffix);
				if ($argn + 1 < @ARGV) {
					$suffix = $ARGV[$argn + 1];
					$argn++;
				} else {
					die "$THIS_FILE: Option requires an argument -- $ARGV[$argn].\n$SHORTHELP\n"
				}
			} elsif ($ARGV[$argn] eq "--help") {
				print $HELPSTR;
				exit;
			} elsif ($ARGV[$argn] eq "--version") {
				print $VERSTR;
				exit;
			} else {
				die "$THIS_FILE: Invalid option -- $ARGV[$argn].\n$SHORTHELP\n";
			}
		} elsif ($ARGV[$argn] =~ /^-/) {
			for ($_ = 1; $_ < length $ARGV[$argn]; $_++) {
				$c = substr ($ARGV[$argn], $_, 1);
				if ($c eq "d") {
					die "$THIS_FILE: -$c declared more than once.\n$SHORTHELP\n"
						if (defined $DEBUG);
					$DEBUG = 1;
				} elsif ($c eq "t") {
					die "$THIS_FILE: -$c declared more than once.\n$SHORTHELP\n"
						if (defined $USETMP);
					$USETMP = 1;
				} elsif ($c eq "s") {
					die "$THIS_FILE: -$c declared more than once.\n$SHORTHELP\n"
						if (defined $suffix);
					if ($_ + 1 < length $ARGV[$argn]) {
						$suffix = substr ($ARGV[$argn], $_ + 1);
						last;
					} elsif ($argn + 1 < @ARGV) {
						$suffix = $ARGV[$argn + 1];
						$argn++;
						last;
					} else {
						die "$THIS_FILE: Option requires an argument -- -$c.\n$SHORTHELP\n"
					}
				} elsif ($c eq "h" || $c eq "?") {
					print $HELPSTR;
					exit;
				} elsif ($c eq "v") {
					print $VERSTR;
					exit;
				} else {
					die "$THIS_FILE: Invalid option -- -$c.\n$SHORTHELP\n";
				}
			}
		} else {
			if (!defined $source) {
				$source = $ARGV[$argn];
			} else {
				die "$THIS_FILE: Too many arguments -- $ARGV[$argn].\n$SHORTHELP\n";
			}
		}
	}
	
	die "$THIS_FILE: Please specify the log file.\n$SHORTHELP\n"
		unless (defined $source);
	
	$source = "$LOG_DIR/$source" if ($source !~ /^\//);
	$source .= "/access_log" if (-d $source);
	die "$THIS_FILE: $source does not exist.\n$SHORTHELP\n"
		unless (-e $source);
	
	$suffix = $DEFAULT_SUFFIX unless (defined $suffix);
	$result = $source . $suffix;
	return ($source, $result);
}
